from __future__ import annotations

import asyncio
import inspect
from typing import Any, Awaitable, Callable, Collection, Sequence, Tuple, Type, TypeVar, cast

__all__ = (
    "AsyncBarrier",
    "cancel_tasks",
    "current_loop",
    "run_through",
)

RT = TypeVar("RT")


async def cancel_tasks(
    tasks: Collection[asyncio.Task[RT]],
) -> Sequence[RT | Exception]:
    """
    Cancel all unfinished tasks from the given collection of asyncio tasks,
    using :func:`asyncio.gather()` to let them clean up concurrently.
    It returns the results and exceptions without raising them, for cases when
    the caller wants to silent ignore errors or handle them at once.
    """
    copied_tasks = {*tasks}
    cancelled_tasks = []
    for task in copied_tasks:
        if not task.done():
            task.cancel()
            cancelled_tasks.append(task)
    return await asyncio.gather(*cancelled_tasks, return_exceptions=True)


current_loop: Callable[[], asyncio.AbstractEventLoop]
if hasattr(asyncio, "get_running_loop"):
    current_loop = asyncio.get_running_loop  # type: ignore
else:
    current_loop = asyncio.get_event_loop  # type: ignore


async def run_through(
    *awaitable_or_callables: Callable[[], None] | Awaitable[None],
    ignored_exceptions: Tuple[Type[Exception], ...],
) -> None:
    """
    A syntactic sugar to simplify the code patterns like:

    .. code-block:: python3

       try:
           await do1()
       except MyError:
           pass
       try:
           await do2()
       except MyError:
           pass
       try:
           await do3()
       except MyError:
           pass

    Using ``run_through()``, it becomes:

    .. code-block:: python3

       await run_through(
           do1(),
           do2(),
           do3(),
           ignored_exceptions=(MyError,),
       )
    """
    for f in awaitable_or_callables:
        try:
            if inspect.iscoroutinefunction(f):
                await f()  # type: ignore
            elif inspect.isawaitable(f):
                await f  # type: ignore
            else:
                f()  # type: ignore
        except Exception as e:
            if isinstance(e, cast(Tuple[Any, ...], ignored_exceptions)):
                continue
            raise


class AsyncBarrier:
    """
    This class provides a simplified asyncio-version of threading.Barrier class.
    """

    num_parties: int = 1
    cond: asyncio.Condition

    def __init__(self, num_parties: int) -> None:
        self.num_parties = num_parties
        self.count = 0
        self.cond = asyncio.Condition()

    async def wait(self) -> None:
        async with self.cond:
            self.count += 1
            if self.count == self.num_parties:
                self.cond.notify_all()
            else:
                while self.count < self.num_parties:
                    await self.cond.wait()

    def reset(self) -> None:
        self.count = 0
        # FIXME: if there are waiting coroutines, let them
        #        raise BrokenBarrierError like threading.Barrier



import ipaddress
import itertools
import json
import logging
import os
import re
import sys
from pathlib import Path
from typing import (
    Any,
    Dict,
    Final,
    Iterable,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import aiohttp
import yarl
from packaging import version

from .etcd import AsyncEtcd
from .etcd import quote as etcd_quote
from .etcd import unquote as etcd_unquote
from .exception import UnknownImageRegistry
from .logging import BraceStyleAdapter

__all__ = (
    "arch_name_aliases",
    "default_registry",
    "default_repository",
    "docker_api_arch_aliases",
    "login",
    "get_known_registries",
    "is_known_registry",
    "get_registry_info",
    "MIN_KERNELSPEC",
    "MAX_KERNELSPEC",
    "ImageRef",
)

arch_name_aliases: Final[Mapping[str, str]] = {
    "arm64": "aarch64",  # macOS with LLVM
    "amd64": "x86_64",  # Windows/Linux
    "x64": "x86_64",  # Windows
    "x32": "x86",  # Windows
    "i686": "x86",  # Windows
}
# generalize architecture symbols to match docker API's norm
docker_api_arch_aliases: Final[Mapping[str, str]] = {
    "aarch64": "arm64",
    "arm64": "arm64",
    "x86_64": "amd64",
    "x64": "amd64",
    "amd64": "amd64",
    "x86": "386",
    "x32": "386",
    "i686": "386",
    "386": "386",
}

log = BraceStyleAdapter(logging.Logger("ai.backend.common.docker"))

default_registry = "index.docker.io"
default_repository = "lablup"

MIN_KERNELSPEC = 1
MAX_KERNELSPEC = 1


def get_docker_connector() -> tuple[yarl.URL, aiohttp.BaseConnector]:
    connector_cls: Type[aiohttp.UnixConnector] | Type[aiohttp.NamedPipeConnector]
    if raw_docker_host := os.environ.get("DOCKER_HOST", None):
        docker_host = yarl.URL(raw_docker_host)
        match docker_host.scheme:
            case "http" | "https":
                return docker_host, aiohttp.TCPConnector()
            case "unix":
                search_paths = [Path(docker_host.path)]
                connector_cls = aiohttp.UnixConnector
            case "npipe":
                search_paths = [Path(docker_host.path.replace("/", "\\"))]
                connector_cls = aiohttp.NamedPipeConnector
            case _ as unknown_scheme:
                raise RuntimeError("Unsupported connection scheme", unknown_scheme)
    else:
        match sys.platform:
            case "linux" | "darwin":
                search_paths = [
                    Path("/run/docker.sock"),
                    Path("/var/run/docker.sock"),
                    Path.home() / ".docker/run/docker.sock",
                ]
                connector_cls = aiohttp.UnixConnector
            case "win32":
                search_paths = [
                    Path(r"\\.\pipe\docker_engine"),
                ]
                connector_cls = aiohttp.NamedPipeConnector
            case _ as unsupported_platform:
                raise RuntimeError("Unsupported platform", unsupported_platform)
    for p in search_paths:
        if p.exists() and (p.is_socket() or p.is_fifo()):
            decoded_path = os.fsdecode(p)
            return (
                yarl.URL("http://localhost"),
                connector_cls(decoded_path),
            )
    else:
        raise RuntimeError("could not find the docker socket")


async def login(
    sess: aiohttp.ClientSession, registry_url: yarl.URL, credentials: dict, scope: str
) -> dict:
    """
    Authorize to the docker registry using the given credentials and token scope, and returns a set
    of required aiohttp.ClientSession.request() keyword arguments for further API requests.

    Some registry servers only rely on HTTP Basic Authentication without token-based access controls
    (usually via nginx proxy). We do support them also. :)
    """
    basic_auth: Optional[aiohttp.BasicAuth]

    if credentials.get("username") and credentials.get("password"):
        basic_auth = aiohttp.BasicAuth(
            credentials["username"],
            credentials["password"],
        )
    else:
        basic_auth = None
    realm = registry_url / "token"  # fallback
    service = "registry"  # fallback
    async with sess.get(registry_url / "v2/", auth=basic_auth) as resp:
        ping_status = resp.status
        www_auth_header = resp.headers.get("WWW-Authenticate")
        if www_auth_header:
            match = re.search(r'realm="([^"]+)"', www_auth_header)
            if match:
                realm = yarl.URL(match.group(1))
            match = re.search(r'service="([^"]+)"', www_auth_header)
            if match:
                service = match.group(1)
    if ping_status == 200:
        log.debug("docker-registry: {0} -> basic-auth", registry_url)
        return {"auth": basic_auth, "headers": {}}
    elif ping_status == 404:
        raise RuntimeError(
            f"Unsupported docker registry: {registry_url}! " "(API v2 not implemented)"
        )
    elif ping_status == 401:
        params = {
            "scope": scope,
            "offline_token": "true",
            "client_id": "docker",
            "service": service,
        }
        async with sess.get(realm, params=params, auth=basic_auth) as resp:
            log.debug("docker-registry: {0} -> {1}", registry_url, realm)
            if resp.status == 200:
                data = json.loads(await resp.read())
                token = data.get("token", None)
                return {
                    "auth": None,
                    "headers": {
                        "Authorization": f"Bearer {token}",
                    },
                }
    raise RuntimeError("authentication for docker registry " f"{registry_url} failed")


async def get_known_registries(etcd: AsyncEtcd) -> Mapping[str, yarl.URL]:
    data = await etcd.get_prefix("config/docker/registry/")
    results: MutableMapping[str, yarl.URL] = {}
    for key, value in data.items():
        name = etcd_unquote(key)
        if isinstance(value, str):
            results[name] = yarl.URL(value)
        elif isinstance(value, Mapping):
            results[name] = yarl.URL(value[""])
    return results


def is_known_registry(val: str, known_registries: Union[Mapping[str, Any], Sequence[str]] = None):
    if val == default_registry:
        return True
    if known_registries is not None and val in known_registries:
        return True
    try:
        url = yarl.URL("//" + val)
        if url.host and ipaddress.ip_address(url.host):
            return True
    except ValueError:
        pass
    return False


async def get_registry_info(etcd: AsyncEtcd, name: str) -> Tuple[yarl.URL, dict]:
    reg_path = f"config/docker/registry/{etcd_quote(name)}"
    item = await etcd.get_prefix(reg_path)
    if not item:
        raise UnknownImageRegistry(name)
    registry_addr = item[""]
    if not registry_addr:
        raise UnknownImageRegistry(name)
    creds = {}
    username = item.get("username")
    if username is not None:
        creds["username"] = username
    password = item.get("password")
    if password is not None:
        creds["password"] = password
    return yarl.URL(registry_addr), creds


class PlatformTagSet(Mapping):

    __slots__ = ("_data",)
    _data: Dict[str, str]
    _rx_ver = re.compile(r"^(?P<tag>[a-zA-Z]+)(?P<version>\d+(?:\.\d+)*[a-z0-9]*)?$")

    def __init__(self, tags: Iterable[str]):
        self._data = dict()
        rx = type(self)._rx_ver
        for t in tags:
            match = rx.search(t)
            if match is None:
                raise ValueError("invalid tag-version string", t)
            key = match.group("tag")
            value = match.group("version")
            if key in self._data:
                raise ValueError("duplicate platform tag with different versions", t)
            if value is None:
                value = ""
            self._data[key] = value

    def has(self, key: str, version: str = None):
        if version is None:
            return key in self._data
        _v = self._data.get(key, None)
        return _v == version

    def __getitem__(self, key: str):
        return self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __eq__(self, other):
        if isinstance(other, (set, frozenset)):
            return set(self._data.keys()) == other
        return self._data == other


class ImageRef:
    """
    Class to represent image reference.
    passing ['*'] to `known_registries` when creating object
    will allow any repository on canonical string.
    """

    __slots__ = ("_registry", "_name", "_tag", "_arch", "_tag_set", "_sha")

    _rx_slug = re.compile(r"^[A-Za-z0-9](?:[A-Za-z0-9-._]*[A-Za-z0-9])?$")

    def __init__(
        self,
        value: str,
        known_registries: Union[Mapping[str, Any], Sequence[str]] = None,
        architecture="x86_64",
    ):
        self._arch = arch_name_aliases.get(architecture, architecture)
        rx_slug = type(self)._rx_slug
        if "://" in value or value.startswith("//"):
            raise ValueError("ImageRef should not contain the protocol scheme.")
        parts = value.split("/", maxsplit=1)
        if len(parts) == 1:
            self._registry = default_registry
            self._name, self._tag = ImageRef._parse_image_tag(value, True)
            if not rx_slug.search(self._tag):
                raise ValueError("Invalid image tag")
        else:
            if is_known_registry(parts[0], known_registries):
                self._registry = parts[0]
                using_default = parts[0].endswith(".docker.io") or parts[0] == "docker.io"
                self._name, self._tag = ImageRef._parse_image_tag(parts[1], using_default)
            # add ['*'] as magic keyword to accept any repository as valid repo
            elif known_registries == ["*"]:
                self._registry = parts[0]
                self._name, self._tag = ImageRef._parse_image_tag(parts[1], False)
            else:
                self._registry = default_registry
                self._name, self._tag = ImageRef._parse_image_tag(value, True)
            if not rx_slug.search(self._tag):
                raise ValueError("Invalid image tag")
        self._update_tag_set()

    @staticmethod
    def _parse_image_tag(s: str, using_default_registry: bool = False) -> Tuple[str, str]:
        image_tag = s.rsplit(":", maxsplit=1)
        if len(image_tag) == 1:
            image = image_tag[0]
            tag = "latest"
        else:
            image = image_tag[0]
            tag = image_tag[1]
        if not image:
            raise ValueError("Empty image repository/name")
        if ("/" not in image) and using_default_registry:
            image = default_repository + "/" + image
        return image, tag

    def _update_tag_set(self):
        if self._tag is None:
            self._tag_set = (None, PlatformTagSet([]))
            return
        tags = self._tag.split("-")
        self._tag_set = (tags[0], PlatformTagSet(tags[1:]))

    def generate_aliases(self) -> Mapping[str, "ImageRef"]:
        basename = self.name.split("/")[-1]
        possible_names = basename.rsplit("-")
        if len(possible_names) > 1:
            possible_names = [basename, possible_names[1]]

        possible_ptags = []
        tag_set = self.tag_set
        if not tag_set[0]:
            pass
        else:
            possible_ptags.append([tag_set[0]])
            for tag_key in tag_set[1]:
                tag_ver = tag_set[1][tag_key]
                tag_list = ["", tag_key, tag_key + tag_ver]
                if "." in tag_ver:
                    tag_list.append(tag_key + tag_ver.rsplit(".")[0])
                elif tag_key == "py" and len(tag_ver) > 1:
                    tag_list.append(tag_key + tag_ver[0])
                if "cuda" in tag_key:
                    tag_list.append("gpu")
                possible_ptags.append(tag_list)

        ret = {}
        for name in possible_names:
            ret[name] = self
        for name, ptags in itertools.product(possible_names, itertools.product(*possible_ptags)):
            ret[f"{name}:{'-'.join(t for t in ptags if t)}"] = self
        return ret

    @staticmethod
    def merge_aliases(genned_aliases_1, genned_aliases_2) -> Mapping[str, "ImageRef"]:
        ret = {}
        aliases_set_1, aliases_set_2 = set(genned_aliases_1.keys()), set(genned_aliases_2.keys())
        aliases_dup = aliases_set_1 & aliases_set_2

        for alias in aliases_dup:
            ret[alias] = max(genned_aliases_1[alias], genned_aliases_2[alias])

        for alias in aliases_set_1 - aliases_dup:
            ret[alias] = genned_aliases_1[alias]
        for alias in aliases_set_2 - aliases_dup:
            ret[alias] = genned_aliases_2[alias]

        return ret

    @property
    def canonical(self) -> str:
        # e.g., registry.docker.io/lablup/kernel-python:3.6-ubuntu
        return f"{self.registry}/{self.name}:{self.tag}"

    @property
    def registry(self) -> str:
        # e.g., lablup
        return self._registry

    @property
    def name(self) -> str:
        # e.g., python
        return self._name

    @property
    def tag(self) -> str:
        # e.g., 3.6-ubuntu
        return self._tag

    @property
    def architecture(self) -> str:
        # e.g., aarch64
        return self._arch

    @property
    def tag_set(self) -> Tuple[str, PlatformTagSet]:
        # e.g., '3.6', {'ubuntu', 'cuda', ...}
        return self._tag_set

    @property
    def short(self) -> str:
        """
        Returns the image reference string without the registry part.
        """
        # e.g., python:3.6-ubuntu
        return f"{self.name}:{self.tag}" if self.tag is not None else self.name

    def __str__(self) -> str:
        return self.canonical

    def __repr__(self) -> str:
        return f'<ImageRef: "{self.canonical}" ({self.architecture})>'

    def __hash__(self) -> int:
        return hash((self._name, self._tag, self._registry, self._arch))

    def __eq__(self, other) -> bool:
        return (
            self._registry == other._registry
            and self._name == other._name
            and self._tag == other._tag
            and self._arch == other._arch
        )

    def __ne__(self, other) -> bool:
        return (
            self._registry != other._registry
            or self._name != other._name
            or self._tag != other._tag
            or self._arch != other._arch
        )

    def __lt__(self, other) -> bool:
        if self == other:  # call __eq__ first for resolved check
            return False
        if self.name != other.name:
            raise ValueError("only the image-refs with same names can be compared.")
        if self.tag_set[0] != other.tag_set[0]:
            return version.parse(self.tag_set[0]) < version.parse(other.tag_set[0])
        ptagset_self, ptagset_other = self.tag_set[1], other.tag_set[1]
        for key_self in ptagset_self:
            if ptagset_other.has(key_self):
                version_self, version_other = ptagset_self.get(key_self), ptagset_other.get(
                    key_self
                )
                if version_self and version_other:
                    parsed_version_self, parsed_version_other = version.parse(
                        version_self
                    ), version.parse(version_other)
                    if parsed_version_self != parsed_version_other:
                        return parsed_version_self < parsed_version_other
        return len(ptagset_self) > len(ptagset_other)



from __future__ import annotations

from pathlib import Path
from typing import Callable

import janus

from .asyncio import current_loop
from .types import Sentinel

__all__ = ("AsyncFileWriter",)


class AsyncFileWriter:
    """
    This class provides a context manager for making sequential async
    writes using janus queue.
    """

    def __init__(
        self,
        target_filename: str | Path,
        access_mode: str,
        encode: Callable[[str], bytes] = None,
        max_chunks: int = None,
    ) -> None:
        if max_chunks is None:
            max_chunks = 0
        self._q: janus.Queue[str | bytes | Sentinel] = janus.Queue(maxsize=max_chunks)
        self._target_filename = target_filename
        self._access_mode = access_mode
        self._binary_mode = "b" in access_mode
        if encode is not None:
            self._encode = encode
        else:
            self._encode = lambda v: v.encode()  # default encoder

    async def __aenter__(self):
        loop = current_loop()
        self._fut = loop.run_in_executor(None, self._write)
        return self

    def _write(self) -> None:
        with open(self._target_filename, self._access_mode) as f:
            while True:
                item = self._q.sync_q.get()
                if item is Sentinel.TOKEN:
                    break
                if self._binary_mode:
                    encoded = self._encode(item) if isinstance(item, str) else item
                    f.write(encoded)
                else:
                    f.write(item)
                self._q.sync_q.task_done()

    async def __aexit__(self, exc_type, exc, tb):
        await self._q.async_q.put(Sentinel.TOKEN)
        try:
            await self._fut
        finally:
            self._q.close()
            await self._q.wait_closed()

    async def write(self, item) -> None:
        await self._q.async_q.put(item)




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/ai/backend/common:dist

from setuptools import setup

setup(**{
    'author': 'Lablup Inc. and contributors',
    'classifiers': [
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
    ],
    'description': 'Backend.AI commons library',
    'install_requires': (
        'PyJWT~=2.0',
        'aiodns>=3.0',
        'aiohttp_sse>=2.0',
        'aiohttp~=3.8.1',
        'aiomonitor-ng~=0.7.0',
        'aiotools~=1.5.9',
        'async_timeout~=4.0',
        'asyncudp>=0.4',
        'attrs>=20.3',
        """backend.ai-plugin==22.09.4
""",
        'click>=7.1.2',
        'coloredlogs~=15.0',
        'etcetra==0.1.13',
        'janus>=0.6.1',
        'msgpack>=1.0.0',
        'netifaces~=0.11.0',
        'packaging>=21.3',
        'python-dateutil>=2.8',
        'python-json-logger>=2.0.1',
        'pyzmq~=23.2.0',
        'redis[hiredis]~=4.3.4',
        'tblib~=1.7',
        'tenacity>=8.0',
        'tomli~=2.0.1',
        'trafaret~=2.1',
        'typeguard~=2.10',
        'types-click',
        'types-python-dateutil',
        'types-redis',
        'typing_extensions~=4.3',
        'yarl>=1.7',
    ),
    'license': 'LGPLv3',
    'long_description': """Backend.AI Commons
==================

[![PyPI release version](https://badge.fury.io/py/backend.ai-common.svg)](https://pypi.org/project/backend.ai-common/)
![Supported Python versions](https://img.shields.io/pypi/pyversions/backend.ai-common.svg)
[![Build Status](https://travis-ci.com/lablup/backend.ai-common.svg?branch=master)](https://travis-ci.com/lablup/backend.ai-common)
[![Gitter](https://badges.gitter.im/lablup/backend.ai-common.svg)](https://gitter.im/lablup/backend.ai-common)

Common utilities library for Backend.AI


## Installation

```console
$ pip install backend.ai-common
```

## For development

```console
$ pip install -U pip setuptools
$ pip install -U -r requirements/dev.txt
```

### Running test suite

```console
$ python -m pytest
```

With the default halfstack setup, you may need to set the environment variable `BACKEND_ETCD_ADDR`
to specify the non-standard etcd service port (e.g., `localhost:8110`).

The tests for `common.redis` module requires availability of local TCP ports 16379, 16380, 16381,
26379, 26380, and 26381 to launch a temporary Redis sentinel cluster via `docker compose`.

In macOS, they require a local `redis-server` executable to be installed, preferably via `brew`,
because `docker compose` in macOS does not support host-mode networking and Redis *cannot* be
configured to use different self IP addresses to announce to the cluster nodes and clients.
""",
    'long_description_content_type': 'text/markdown',
    'name': 'backend.ai-common',
    'namespace_packages': (
    ),
    'package_data': {
        'ai.backend.common': (
            'VERSION',
            'enum_extension.pyi',
            'py.typed',
        ),
        'ai.backend.common.plugin': (
            'py.typed',
        ),
    },
    'packages': (
        'ai.backend.common',
        'ai.backend.common.plugin',
        'ai.backend.common.web.session',
    ),
    'project_urls': {
        'Documentation': 'https://docs.backend.ai/',
        'Source': 'https://github.com/lablup/backend.ai',
    },
    'python_requires': '>=3.10,<3.11',
    'url': 'https://github.com/lablup/backend.ai',
    'version': """22.09.4
""",
    'zip_safe': False,
})



import re
from typing import List, Sequence, Set, Type

from .types import ServicePort, ServicePortProtocols

__all__ = "parse_service_ports"

_rx_service_ports = re.compile(
    r"^(?P<name>[\w-]+):(?P<proto>\w+):(?P<ports>\[\d+(?:,\d+)*\]|\d+)(?:,|$)"
)


def parse_service_ports(s: str, exception_cls: Type[Exception] = None) -> Sequence[ServicePort]:
    items: List[ServicePort] = []
    if exception_cls is None:
        exception_cls = ValueError
    used_ports: Set[int] = set()
    while True:
        match = _rx_service_ports.search(s)
        if match:
            s = s[len(match.group(0)) :]
            name = match.group("name")
            if not name:
                raise exception_cls("Service port name must be not empty.")
            protocol = match.group("proto")
            if protocol == "pty":
                # unsupported, skip
                continue
            if protocol not in ("tcp", "http", "preopen"):
                raise exception_cls(f"Unsupported service port protocol: {protocol}")
            ports = tuple(map(int, match.group("ports").strip("[]").split(",")))
            for p in ports:
                if p in used_ports:
                    raise exception_cls(f"The port {p} is already used by another service port.")
                if p <= 1024:
                    raise exception_cls(
                        f"The service port number {p} must be "
                        f"larger than 1024 to run without the root privilege."
                    )
                if p >= 65535:
                    raise exception_cls(f"The service port number {p} must be smaller than 65535.")
                if p in (2000, 2001, 2002, 2003, 2200, 7681):
                    raise exception_cls(
                        "The service ports 2000 to 2003, 2200 and 7681 "
                        "are reserved for internal use."
                    )
                used_ports.add(p)
            items.append(
                {
                    "name": name,
                    "protocol": ServicePortProtocols(protocol),
                    "container_ports": ports,
                    "host_ports": (None,) * len(ports),
                }
            )
        else:
            if len(s) > 0:
                raise exception_cls("Invalid format")
            break
    return items
